@@ -18,6 +18,7 @@
 """Model subclass to wrap scikit-learn SGD classifier and regressor models."""
 
 import typing
+import warnings
 from typing import Any, Callable, Dict, Literal, Optional, Set, Tuple, Union
 
 import numpy as np
@@ -29,7 +30,7 @@ from declearn.data_info import aggregate_data_info
 from declearn.model.api import Model
 from declearn.model.sklearn._np_vec import NumpyVector
 from declearn.typing import Batch
-from declearn.utils import register_type
+from declearn.utils import DevicePolicy, register_type
 
 
 __all__ = [
@@ -63,6 +64,13 @@ class SklearnSGDModel(Model):
     This `Model` subclass is designed to wrap a `SGDClassifier`
     or `SGDRegressor` instance (from `sklearn.linear_model`) to
     be learned federatively.
+
+    Notes regarding device management (CPU, GPU, etc.):
+    * This Model may only run on CPU, and is unaffected by device-
+      management policies.
+    * Calling the `update_device_policy` method has no effect, and
+      raises a UserWarning if a GPU-targetting policy is passed to
+      it directly.
     """
 
     def __init__(
@@ -104,6 +112,12 @@ class SklearnSGDModel(Model):
             None
         )  # type: Optional[Callable[[np.ndarray, np.ndarray], np.ndarray]]
 
+    @property
+    def device_policy(
+        self,
+    ) -> DevicePolicy:
+        return DevicePolicy(gpu=False, idx=None)
+
     @property
     def required_data_info(
         self,
@@ -137,7 +151,7 @@ class SklearnSGDModel(Model):
         cls,
         kind: Literal["classifier", "regressor"],
         loss: Optional[LossesLiteral] = None,
-        penalty: Literal["l1", "l2", "elasticnet"] = "l2",
+        penalty: Literal["none", "l1", "l2", "elasticnet"] = "l2",
         alpha: float = 1e-4,
         l1_ratio: float = 0.15,
         epsilon: float = 0.1,
@@ -159,7 +173,7 @@ class SklearnSGDModel(Model):
             See `sklearn.linear_model.SGDRegressor` and `SGDClassifier`
             documentation for details on possible values. If None, set
             to "hinge" for classifier or "squared_error" for regressor.
-        penalty: {"l1", "l2", "elasticnet"}, default="l2"
+        penalty: {"none", "l1", "l2", "elasticnet"}, default="l2"
             The penalty (i.e. regularization term) to be used.
         alpha: float, default=0.0001
             Regularization constant (the higher the stronger).
@@ -252,6 +266,7 @@ class SklearnSGDModel(Model):
 
     def get_weights(
         self,
+        trainable: bool = False,
     ) -> NumpyVector:
         weights = {
             "intercept": self._model.intercept_.copy(),
@@ -262,12 +277,13 @@ class SklearnSGDModel(Model):
     def set_weights(  # type: ignore  # Vector subtype specification
         self,
         weights: NumpyVector,
+        trainable: bool = False,
     ) -> None:
         if not isinstance(weights, NumpyVector):
             raise TypeError("SklearnSGDModel requires NumpyVector weights.")
         for key in ("coef", "intercept"):
             if key not in weights.coefs:
-                raise TypeError(
+                raise KeyError(
                     f"Missing required '{key}' in the received vector."
                 )
         self._model.coef_ = weights.coefs["coef"].copy()
@@ -382,3 +398,10 @@ class SklearnSGDModel(Model):
         else:
             loss_fn = loss_1d
         return loss_fn
+
+    def update_device_policy(
+        self,
+        policy: Optional[DevicePolicy] = None,
+    ) -> None:
+        if policy is not None and policy.gpu:
+            warnings.warn("'SklearnSGDModel' only runs on a CPU backend.")
