@@ -17,7 +17,6 @@
 
 """Model subclass to wrap TensorFlow models."""
 
-import warnings
 from copy import deepcopy
 from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
 
@@ -28,18 +27,44 @@ from typing_extensions import Self  # future: import from typing (py >=3.11)
 
 from declearn.data_info import aggregate_data_info
 from declearn.model.api import Model
-from declearn.model.tensorflow._utils import build_keras_loss
+from declearn.model.tensorflow.utils import (
+    build_keras_loss,
+    move_layer_to_device,
+    select_device,
+)
 from declearn.model.tensorflow._vector import TensorflowVector
+from declearn.model._utils import raise_on_stringsets_mismatch
 from declearn.typing import Batch
-from declearn.utils import register_type
+from declearn.utils import DevicePolicy, get_device_policy, register_type
+
+
+__all__ = [
+    "TensorflowModel",
+]
 
 
 @register_type(name="TensorflowModel", group="Model")
 class TensorflowModel(Model):
     """Model wrapper for TensorFlow Model instances.
 
-    This `Model` subclass is designed to wrap a `tf.keras.Model`
-    instance to be learned federatively.
+    This `Model` subclass is designed to wrap a `tf.keras.Model` instance
+    to be trained federatively.
+
+    Notes regarding device management (CPU, GPU, etc.):
+    * By default, tensorflow places data and operations on GPU whenever one
+      is available.
+    * Our `TensorflowModel` instead consults the device-placement policy (via
+      `declearn.utils.get_device_policy`), places the wrapped keras model's
+      weights there, and runs computations defined under public methods in
+      a `tensorflow.device` context, to enforce that policy.
+    * Note that there is no guarantee that calling a private method directly
+      will result in abiding by that policy. Hence, be careful when writing
+      custom code, and use your own context managers to get guarantees.
+    * Note that if the global device-placement policy is updated, this will
+      only be propagated to existing instances by manually calling their
+      `update_device_policy` method.
+    * You may consult the device policy currently enforced by a TensorflowModel
+      instance by accessing its `device_policy` property.
     """
 
     def __init__(
@@ -47,6 +72,7 @@ class TensorflowModel(Model):
         model: tf.keras.layers.Layer,
         loss: Union[str, tf.keras.losses.Loss],
         metrics: Optional[List[Union[str, tf.keras.metrics.Metric]]] = None,
+        _from_config: bool = False,
         **kwargs: Any,
     ) -> None:
         """Instantiate a Model interface wrapping a tensorflow.keras model.
@@ -66,7 +92,7 @@ class TensorflowModel(Model):
             compiled with the model and computed using the `evaluate`
             method of the returned TensorflowModel instance.
         **kwargs: Any
-            Any addition keyword argument to `tf.keras.Model.compile`
+            Any additional keyword argument to `tf.keras.Model.compile`
             may be passed.
         """
         # Type-check the input Model and wrap it up.
@@ -79,12 +105,28 @@ class TensorflowModel(Model):
         super().__init__(model)
         # Ensure the loss is a keras.Loss object and set its reduction to none.
         loss = build_keras_loss(loss, reduction=tf.keras.losses.Reduction.NONE)
+        # Select the device where to place computations and move the model.
+        policy = get_device_policy()
+        self._device = select_device(gpu=policy.gpu, idx=policy.idx)
+        if not _from_config:
+            self._model = move_layer_to_device(self._model, self._device)
+        # Finalize initialization using the selected device.
         # Compile the wrapped model and retain compilation arguments.
-        kwargs.update({"loss": loss, "metrics": metrics})
-        model.compile(**kwargs)
-        self._kwargs = kwargs
-        # Instantiate a SGD optimizer to apply updates as-provided.
-        self._sgd = tf.keras.optimizers.SGD(learning_rate=1.0)
+        with tf.device(self._device):
+            kwargs.update({"loss": loss, "metrics": metrics})
+            self._model.compile(**kwargs)
+            self._kwargs = kwargs
+
+    @property
+    def device_policy(
+        self,
+    ) -> DevicePolicy:
+        device = self._device
+        try:
+            idx = int(device.name.rsplit(":", 1)[-1])
+        except ValueError:
+            idx = None
+        return DevicePolicy(gpu=(device.device_type == "GPU"), idx=idx)
 
     @property
     def required_data_info(
@@ -98,19 +140,8 @@ class TensorflowModel(Model):
     ) -> None:
         if not self._model.built:
             data_info = aggregate_data_info([data_info], {"input_shape"})
-            self._model.build(data_info["input_shape"])
-        # Warn about frozen weights.
-        # similar to TorchModel warning; pylint: disable=duplicate-code
-        if len(self._model.trainable_weights) < len(self._model.weights):
-            warnings.warn(
-                "'TensorflowModel' wraps a model with frozen weights.\n"
-                "This is not fully compatible with declearn v2.0.x: the "
-                "use of weight decay and/or of a loss-regularization "
-                "plug-in in an Optimizer will fail to produce updates "
-                "for this model.\n"
-                "This issue will be fixed in declearn v2.1.0."
-            )
-        # pylint: enable=duplicate-code
+            with tf.device(self._device):
+                self._model.build(data_info["input_shape"])
 
     def get_config(
         self,
@@ -129,35 +160,44 @@ class TensorflowModel(Model):
         for key in ("model", "loss", "kwargs"):
             if key not in config.keys():
                 raise KeyError(f"Missing key '{key}' in the config dict.")
-        model = tf.keras.layers.deserialize(config["model"])
-        loss = tf.keras.losses.deserialize(config["loss"])
-        return cls(model, loss, **config["kwargs"])
+        # Set up the device policy.
+        policy = get_device_policy()
+        device = select_device(gpu=policy.gpu, idx=policy.idx)
+        # Deserialize the model and loss keras objects on the device.
+        with tf.device(device):
+            model = tf.keras.layers.deserialize(config["model"])
+            loss = tf.keras.losses.deserialize(config["loss"])
+        # Instantiate the TensorflowModel, avoiding device-to-device copies.
+        return cls(model, loss, **config["kwargs"], _from_config=True)
 
     def get_weights(
         self,
+        trainable: bool = False,
     ) -> TensorflowVector:
-        # REVISE: only return trainable weights? add flag to select?
-        return TensorflowVector(
-            {var.name: var.value() for var in self._model.weights}
+        variables = (
+            self._model.trainable_weights if trainable else self._model.weights
         )
+        return TensorflowVector({var.name: var.value() for var in variables})
 
     def set_weights(  # type: ignore  # Vector subtype specification
         self,
         weights: TensorflowVector,
+        trainable: bool = False,
     ) -> None:
         if not isinstance(weights, TensorflowVector):
             raise TypeError(
                 "TensorflowModel requires TensorflowVector weights."
             )
-        self._verify_weights_compatibility(weights, trainable_only=False)
+        self._verify_weights_compatibility(weights, trainable=trainable)
         variables = {var.name: var for var in self._model.weights}
-        for name, value in weights.coefs.items():
-            variables[name].assign(value, read_value=False)
+        with tf.device(self._device):
+            for name, value in weights.coefs.items():
+                variables[name].assign(value, read_value=False)
 
     def _verify_weights_compatibility(
         self,
         vector: TensorflowVector,
-        trainable_only: bool = False,
+        trainable: bool = False,
     ) -> None:
         """Verify that a vector has the same names as the model's weights.
 
@@ -166,7 +206,7 @@ class TensorflowModel(Model):
         vector: TensorflowVector
             Vector wrapping weight-related coefficients (e.g. weight
             values or gradient-based updates).
-        trainable_only: bool, default=False
+        trainable: bool, default=False
             Whether to restrict the comparision to the model's trainable
             weights rather than to all of its weights.
 
@@ -176,33 +216,27 @@ class TensorflowModel(Model):
             In case some expected keys are missing, or additional keys
             are present. Be verbose about the identified mismatch(es).
         """
-        # Gather the variables to compare to the input vector.
-        if trainable_only:
-            weights = self._model.trainable_weights
-        else:
-            weights = self._model.weights
-        variables = {var.name: var for var in weights}
-        # Raise a verbose KeyError in case inputs do not match weights.
-        if set(vector.coefs).symmetric_difference(variables):
-            missing = set(variables).difference(vector.coefs)
-            unexpct = set(vector.coefs).difference(variables)
-            raise KeyError(
-                "Mismatch between input and model weights' names:\n"
-                + f"Missing key(s) in inputs: {missing}\n" * bool(missing)
-                + f"Unexpected key(s) in inputs: {unexpct}\n" * bool(unexpct)
-            )
+        variables = (
+            self._model.trainable_weights if trainable else self._model.weights
+        )
+        raise_on_stringsets_mismatch(
+            received=set(vector.coefs),
+            expected={var.name for var in variables},
+            context="model weights",
+        )
 
     def compute_batch_gradients(
         self,
         batch: Batch,
         max_norm: Optional[float] = None,
     ) -> TensorflowVector:
-        data = self._unpack_batch(batch)
-        if max_norm is None:
-            grads = self._compute_batch_gradients(*data)
-        else:
-            norm = tf.constant(max_norm)
-            grads = self._compute_clipped_gradients(*data, norm)
+        with tf.device(self._device):
+            data = self._unpack_batch(batch)
+            if max_norm is None:
+                grads = self._compute_batch_gradients(*data)
+            else:
+                norm = tf.constant(max_norm)
+                grads = self._compute_clipped_gradients(*data, norm)
         grads_and_vars = zip(grads, self._model.trainable_weights)
         return TensorflowVector(
             {var.name: grad for grad, var in grads_and_vars}
@@ -284,14 +318,14 @@ class TensorflowModel(Model):
         self,
         updates: TensorflowVector,
     ) -> None:
-        self._verify_weights_compatibility(updates, trainable_only=True)
-        # Delegate updates' application to a tensorflow Optimizer.
-        values = (-1 * updates).coefs.values()
-        zipped = zip(values, self._model.trainable_weights)
-        upd_op = self._sgd.apply_gradients(zipped)
-        # Ensure ops have been performed before exiting.
-        with tf.control_dependencies([upd_op]):
-            return None
+        self._verify_weights_compatibility(updates, trainable=True)
+        with tf.device(self._device):
+            for var in self._model.trainable_weights:
+                updt = updates.coefs[var.name]
+                if isinstance(updt, tf.IndexedSlices):
+                    var.scatter_add(updt)
+                else:
+                    var.assign_add(updt, read_value=False)
 
     def evaluate(
         self,
@@ -311,21 +345,23 @@ class TensorflowModel(Model):
         metrics: dict[str, float]
             Dictionary associating evaluation metrics' values to their name.
         """
-        return self._model.evaluate(dataset, return_dict=True)
+        with tf.device(self._device):
+            return self._model.evaluate(dataset, return_dict=True)
 
     def compute_batch_predictions(
         self,
         batch: Batch,
     ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
-        inputs, y_true, s_wght = self._unpack_batch(batch)
-        if y_true is None:
-            raise TypeError(
-                "`TensorflowModel.compute_batch_predictions` received a "
-                "batch with `y_true=None`, which is unsupported. Please "
-                "correct the inputs, or override this method to support "
-                "creating labels from the base inputs."
-            )
-        y_pred = self._model(inputs, training=False).numpy()
+        with tf.device(self._device):
+            inputs, y_true, s_wght = self._unpack_batch(batch)
+            if y_true is None:
+                raise TypeError(
+                    "`TensorflowModel.compute_batch_predictions` received a "
+                    "batch with `y_true=None`, which is unsupported. Please "
+                    "correct the inputs, or override this method to support "
+                    "creating labels from the base inputs."
+                )
+            y_pred = self._model(inputs, training=False).numpy()
         y_true = y_true.numpy()
         s_wght = s_wght.numpy() if s_wght is not None else s_wght
         return y_true, y_pred, s_wght
@@ -335,7 +371,21 @@ class TensorflowModel(Model):
         y_true: np.ndarray,
         y_pred: np.ndarray,
     ) -> np.ndarray:
-        tns_true = tf.convert_to_tensor(y_true)
-        tns_pred = tf.convert_to_tensor(y_pred)
-        s_loss = self._model.compute_loss(y=tns_true, y_pred=tns_pred)
+        with tf.device(self._device):
+            tns_true = tf.convert_to_tensor(y_true)
+            tns_pred = tf.convert_to_tensor(y_pred)
+            s_loss = self._model.compute_loss(y=tns_true, y_pred=tns_pred)
         return s_loss.numpy().squeeze()
+
+    def update_device_policy(
+        self,
+        policy: Optional[DevicePolicy] = None,
+    ) -> None:
+        # Select the device to use based on the provided or global policy.
+        if policy is None:
+            policy = get_device_policy()
+        device = select_device(gpu=policy.gpu, idx=policy.idx)
+        # When needed, re-create the model to force moving it to the device.
+        if self._device is not device:
+            self._device = device
+            self._model = move_layer_to_device(self._model, self._device)
