@@ -27,9 +27,16 @@ import torch
 from typing_extensions import Self  # future: import from typing (py >=3.11)
 
 from declearn.model.api import Model
+from declearn.model.torch.utils import AutoDeviceModule, select_device
 from declearn.model.torch._vector import TorchVector
+from declearn.model._utils import raise_on_stringsets_mismatch
 from declearn.typing import Batch
-from declearn.utils import register_type
+from declearn.utils import DevicePolicy, get_device_policy, register_type
+
+
+__all__ = [
+    "TorchModel",
+]
 
 
 # alias for unpacked Batch structures, converted to torch.Tensor objects
@@ -42,8 +49,23 @@ TensorBatch = Tuple[
 class TorchModel(Model):
     """Model wrapper for PyTorch Model instances.
 
-    This `Model` subclass is designed to wrap a `torch.nn.Module`
-    instance to be learned federatively.
+    This `Model` subclass is designed to wrap a `torch.nn.Module` instance
+    to be trained federatively.
+
+    Notes regarding device management (CPU, GPU, etc.):
+    * By default torch operates on CPU, and it does not automatically move
+      tensors between devices. This means users have to be careful where
+      tensors are placed to avoid operations between tensors on different
+      devices, leading to runtime errors.
+    * Our `TorchModel` instead consults the global device-placement policy
+      (via `declearn.utils.get_device_policy`), places the wrapped torch
+      modules' weights there, and automates the placement of input data on
+      the same device as the wrapped model.
+    * Note that if the global device-placement policy is updated, this will
+      only be propagated to existing instances by manually calling their
+      `update_device_policy` method.
+    * You may consult the device policy currently enforced by a TorchModel
+      instance by accessing its `device_policy` property.
     """
 
     def __init__(
@@ -62,18 +84,29 @@ class TorchModel(Model):
             is to be minimized through training. Note that it will be
             altered when wrapped.
         """
-        # Type-check the input Model and wrap it up.
+        # Type-check the input model.
         if not isinstance(model, torch.nn.Module):
             raise TypeError("'model' should be a torch.nn.Module instance.")
+        # Select the device where to place computations, and wrap the model.
+        policy = get_device_policy()
+        device = select_device(gpu=policy.gpu, idx=policy.idx)
+        model = AutoDeviceModule(model, device=device)
         super().__init__(model)
         # Assign loss module and set it not to reduce sample-wise values.
         if not isinstance(loss, torch.nn.Module):
             raise TypeError("'loss' should be a torch.nn.Module instance.")
-        self._loss_fn = loss
-        self._loss_fn.reduction = "none"  # type: ignore
+        loss.reduction = "none"  # type: ignore
+        self._loss_fn = AutoDeviceModule(loss, device=device)
         # Compute and assign a functional version of the model.
         self._func_model = functorch.make_functional(self._model)[0]
 
+    @property
+    def device_policy(
+        self,
+    ) -> DevicePolicy:
+        device = self._model.device
+        return DevicePolicy(gpu=(device.type == "cuda"), idx=device.index)
+
     @property
     def required_data_info(
         self,
@@ -84,16 +117,7 @@ class TorchModel(Model):
         self,
         data_info: Dict[str, Any],
     ) -> None:
-        # Warn about frozen weights.
-        if not all(p.requires_grad for p in self._model.parameters()):
-            warnings.warn(
-                "'TorchModel' wraps a model with frozen weights.\n"
-                "This is not fully compatible with declearn v2.0.x: the "
-                "use of weight decay and/or of a loss-regularization "
-                "plug-in in an Optimizer will fail to produce updates "
-                "for this model.\n"
-                "This issue will be fixed in declearn v2.1.0."
-            )
+        return None
 
     def get_config(
         self,
@@ -102,15 +126,12 @@ class TorchModel(Model):
             "PyTorch JSON serialization relies on pickle, which may be unsafe."
         )
         with io.BytesIO() as buffer:
-            torch.save(self._model, buffer)
+            torch.save(self._model.module, buffer)
             model = buffer.getbuffer().hex()
         with io.BytesIO() as buffer:
-            torch.save(self._loss_fn, buffer)
+            torch.save(self._loss_fn.module, buffer)
             loss = buffer.getbuffer().hex()
-        return {
-            "model": model,
-            "loss": loss,
-        }
+        return {"model": model, "loss": loss}
 
     @classmethod
     def from_config(
@@ -126,20 +147,63 @@ class TorchModel(Model):
 
     def get_weights(
         self,
+        trainable: bool = False,
     ) -> TorchVector:
-        weights = {
-            key: tns.detach().clone()  # NOTE: otherwise, view on Tensor
-            for key, tns in self._model.state_dict().items()
-        }
-        return TorchVector(weights)
+        params = self._model.named_parameters()
+        if trainable:
+            weights = {k: p.data for k, p in params if p.requires_grad}
+        else:
+            weights = {k: p.data for k, p in params}
+        # Note: calling `tensor.clone()` to return a copy rather than a view.
+        return TorchVector({k: t.detach().clone() for k, t in weights.items()})
 
     def set_weights(  # type: ignore  # Vector subtype specification
         self,
         weights: TorchVector,
+        trainable: bool = False,
     ) -> None:
         if not isinstance(weights, TorchVector):
             raise TypeError("TorchModel requires TorchVector weights.")
-        self._model.load_state_dict(weights.coefs)
+        self._verify_weights_compatibility(weights, trainable=trainable)
+        if trainable:
+            state_dict = self._model.state_dict()
+            state_dict.update(weights.coefs)
+        else:
+            state_dict = weights.coefs
+        # NOTE: this preserves the device placement of current states
+        self._model.load_state_dict(state_dict)
+
+    def _verify_weights_compatibility(
+        self,
+        vector: TorchVector,
+        trainable: bool = False,
+    ) -> None:
+        """Verify that a vector has the same names as the model's weights.
+
+        Parameters
+        ----------
+        vector: TorchVector
+            Vector wrapping weight-related coefficients (e.g. weight
+            values or gradient-based updates).
+        trainable: bool, default=False
+            Whether to restrict the comparision to the model's trainable
+            weights rather than to all of its weights.
+
+        Raises
+        ------
+        KeyError:
+            In case some expected keys are missing, or additional keys
+            are present. Be verbose about the identified mismatch(es).
+        """
+        received = set(vector.coefs)
+        expected = {
+            name
+            for name, param in self._model.named_parameters()
+            if (not trainable) or param.requires_grad
+        }
+        raise_on_stringsets_mismatch(
+            received, expected, context="model weights"
+        )
 
     def compute_batch_gradients(
         self,
@@ -197,7 +261,7 @@ class TorchModel(Model):
         """Compute the average (opt. weighted) loss over given predictions."""
         loss = self._loss_fn(y_pred, y_true)
         if s_wght is not None:
-            loss.mul_(s_wght)
+            loss.mul_(s_wght.to(loss.device))
         return loss.mean()
 
     def _compute_samplewise_gradients(
@@ -236,7 +300,7 @@ class TorchModel(Model):
                 # false-positive; pylint: disable=no-member
                 grad.mul_(torch.clamp(max_norm / norm, max=1))
                 if s_wght is not None:
-                    grad.mul_(s_wght)
+                    grad.mul_(s_wght.to(grad.device))
             return grads
         # Vectorize the function to compute sample-wise clipped gradients.
         with torch.no_grad():
@@ -318,15 +382,11 @@ class TorchModel(Model):
     ) -> None:
         if not isinstance(updates, TorchVector):
             raise TypeError("TorchModel requires TorchVector updates.")
+        self._verify_weights_compatibility(updates, trainable=True)
         with torch.no_grad():
-            try:
-                for key, upd in updates.coefs.items():
-                    tns = self._model.get_parameter(key)
-                    tns.add_(upd)
-            except KeyError as exc:
-                raise KeyError(
-                    "Invalid model parameter name(s) found in updates."
-                ) from exc
+            for key, upd in updates.coefs.items():
+                tns = self._model.get_parameter(key)
+                tns.add_(upd.to(tns.device))
 
     def compute_batch_predictions(
         self,
@@ -342,9 +402,9 @@ class TorchModel(Model):
             )
         self._model.eval()
         with torch.no_grad():
-            y_pred = self._model(*inputs).numpy()
-        y_true = y_true.numpy()
-        s_wght = s_wght.numpy() if s_wght is not None else s_wght
+            y_pred = self._model(*inputs).cpu().numpy()
+        y_true = y_true.cpu().numpy()
+        s_wght = None if s_wght is None else s_wght.cpu().numpy()
         return y_true, y_pred, s_wght  # type: ignore
 
     def loss_function(
@@ -355,4 +415,16 @@ class TorchModel(Model):
         tns_pred = torch.from_numpy(y_pred)  # pylint: disable=no-member
         tns_true = torch.from_numpy(y_true)  # pylint: disable=no-member
         s_loss = self._loss_fn(tns_pred, tns_true)
-        return s_loss.numpy().squeeze()
+        return s_loss.cpu().numpy().squeeze()
+
+    def update_device_policy(
+        self,
+        policy: Optional[DevicePolicy] = None,
+    ) -> None:
+        # Select the device to use based on the provided or global policy.
+        if policy is None:
+            policy = get_device_policy()
+        device = select_device(gpu=policy.gpu, idx=policy.idx)
+        # Place the wrapped model and loss function modules on that device.
+        self._model.set_device(device)
+        self._loss_fn.set_device(device)
