@@ -25,19 +25,47 @@ from typing_extensions import Self  # future: import from typing (Py>=3.11)
 
 from declearn.model.api import Vector, register_vector_type
 from declearn.model.sklearn import NumpyVector
+from declearn.model.torch.utils import select_device
+from declearn.utils import get_device_policy
+
+
+__all__ = [
+    "TorchVector",
+]
 
 
 @register_vector_type(torch.Tensor)
 class TorchVector(Vector):
     """Vector subclass to store PyTorch tensors.
 
-    This Vector is designed to store a collection of named
-    PyTorch tensors, enabling computations that are either
-    applied to each and every coefficient, or imply two sets
-    of aligned coefficients (i.e. two TorchVector with
-    similar specifications).
+    This Vector is designed to store a collection of named PyTorch
+    tensors, enabling computations that are either applied to each
+    and every coefficient, or imply two sets of aligned coefficients
+    (i.e. two TorchVector with similar specifications).
 
     Use `vector.coefs` to access the stored coefficients.
+
+    Notes
+    -----
+    - A `TorchVector` can be operated with either a:
+      - scalar value
+      - `NumpyVector` that has similar specifications
+      - `TorchVector` that has similar specifications
+      => resulting in a `TorchVector` in each of these cases.
+    - The wrapped tensors may be placed on any device (CPU, GPU...)
+      and may not be all on the same device.
+    - The device-placement of the initial `TorchVector`'s data
+      is preserved by operations, including with `NumpyVector`.
+    - When combining two `TorchVector`, the device-placement
+      of the left-most one is used; in that case, one ends up with
+      `gpu + cpu = gpu` while `cpu + gpu = cpu`. In both cases, a
+      warning will be emitted to prevent silent un-optimized copies.
+    - When deserializing a `TorchVector` (either by directly using
+      `TorchVector.unpack` or loading one from a JSON dump), loaded
+      tensors are placed based on the global device-placement policy
+      (accessed via `declearn.utils.get_device_policy`). Thus it may
+      have a different device-placement schema than at dump time but
+      should be coherent with that of `TorchModel` computations.
     """
 
     @property
@@ -73,12 +101,20 @@ class TorchVector(Vector):
         other: Any,
         func: Callable[[Any, Any], Any],
     ) -> Self:
+        # Convert 'other' NumpyVector to a (CPU-backed) TorchVector.
         if isinstance(other, NumpyVector):
             # false-positive; pylint: disable=no-member
             coefs = {
                 key: torch.from_numpy(val) for key, val in other.coefs.items()
             }
             other = TorchVector(coefs)
+        # Ensure 'other' TorchVector shares this vector's device placement.
+        if isinstance(other, TorchVector):
+            coefs = {
+                key: val.to(self.coefs[key].device)
+                for key, val in other.coefs.items()
+            }
+            other = TorchVector(coefs)
         return super()._apply_operation(other, func)
 
     def dtypes(
@@ -95,15 +131,22 @@ class TorchVector(Vector):
     def pack(
         self,
     ) -> Dict[str, Any]:
-        return {key: tns.numpy() for key, tns in self.coefs.items()}
+        return {
+            key: np.array(tns.cpu().numpy()) for key, tns in self.coefs.items()
+        }
 
     @classmethod
     def unpack(
         cls,
         data: Dict[str, Any],
     ) -> Self:
-        # false-positive; pylint: disable=no-member
-        coefs = {key: torch.from_numpy(dat) for key, dat in data.items()}
+        policy = get_device_policy()
+        device = select_device(gpu=policy.gpu, idx=policy.idx)
+        coefs = {
+            # false-positive on `torch.from_numpy`; pylint: disable=no-member
+            key: torch.from_numpy(dat).to(device)
+            for key, dat in data.items()
+        }
         return cls(coefs)
 
     def __eq__(
@@ -115,8 +158,9 @@ class TorchVector(Vector):
             valid = self.coefs.keys() == other.coefs.keys()
         if valid:
             valid = all(
-                np.array_equal(self.coefs[k].numpy(), other.coefs[k].numpy())
-                for k in self.coefs
+                # false-positive on 'torch.equal'; pylint: disable=no-member
+                torch.equal(tns, other.coefs[key].to(tns.device))
+                for key, tns in self.coefs.items()
             )
         return valid
 
