@@ -17,7 +17,11 @@
 
 """Collection of utils for running tests and examples around declearn."""
 
-from ._assertions import assert_json_serializable_dict
+from ._assertions import (
+    assert_dict_equal,
+    assert_list_equal,
+    assert_json_serializable_dict,
+)
 from ._gen_ssl import generate_ssl_certificates
 from ._multiprocess import run_as_processes
 from ._vectors import (
